%{ /* -*- makefile -*- */
/* 
 * We build a lexical analyzer that converts a Java source file to 
 * a beautifully highlighted HTML file now.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "colors.h"

#ifndef COMPRESSION
#define COMPRESSION 0
#endif

FILE *actin;
FILE *actout;

char *ownhead = "head.java2html";
char *ownbottom = "bottom.java2html";

unsigned int oldstate = 0;

%}
%s COMMENT
%s ONELINECOMMENT
%s STRING
 /* We define handling for #define etc to support C++ with this
    program too */
%s DEFINELINE
%s PREPROLINE

id [a-zA-Z_][a-zA-Z01-9_]*
wsnl [\t \n]*
ws   [\t ]*
accessmode ("public"|"private"|"protected")
 /* when to break parsing comments, strings, etc to handle special 
    characters: */
commentnospecial         [^&<>----\\\n*/]
onelinecommentnospecial [^&<>----\n]
stringnospecial          [^&<>----\\\n"]
definelinenospecial      [^&<>----\\\n"*/]
preprolinenospecial      [^&<>----\n"*/]

%%
	/* This characters have to be recoded everywhere
	 * because the WWW-browsers interpret them in their
	 * own (here not wanted) way.
	 */
"&"	{ fprintf (actout, "&amp;"); }
"<"	{ fprintf (actout, "&lt;"); }
">"	{ fprintf (actout, "&gt;"); }
	/* All character as latin1 greater than dec192 or hexBF */
""	{ fprintf (actout, "&Agrave;"); }
""	{ fprintf (actout, "&Aacute;"); }
""	{ fprintf (actout, "&Acirc;"); }
""	{ fprintf (actout, "&Atilde;"); }
""	{ fprintf (actout, "&Auml;"); } 
""	{ fprintf (actout, "&Aring;"); }
""	{ fprintf (actout, "&AElig;"); }
""	{ fprintf (actout, "&Ccedil;"); }
""	{ fprintf (actout, "&Egrave;"); }
""	{ fprintf (actout, "&Eacute;"); }
""	{ fprintf (actout, "&Ecirc;"); } 
""	{ fprintf (actout, "&Euml;"); }  
""	{ fprintf (actout, "&Igrave;"); }
""	{ fprintf (actout, "&Iacute;"); }
""	{ fprintf (actout, "&Icirc;"); } 
""	{ fprintf (actout, "&Iuml;"); }  
	/* ""	{ fprintf (actout, "", yytext); } */
""	{ fprintf (actout, "&Ntilde;"); }
""	{ fprintf (actout, "&Ograve;"); }
""	{ fprintf (actout, "&Oacute;"); }
""	{ fprintf (actout, "&Ocirc;"); } 
""	{ fprintf (actout, "&Otilde;"); }
""	{ fprintf (actout, "&Ouml;"); }  
	/* ""	{ fprintf (actout, "", yytext); } */
""	{ fprintf (actout, "&Oslash;"); }
""	{ fprintf (actout, "&Ugrave;"); }
""	{ fprintf (actout, "&Uacute;"); }
""	{ fprintf (actout, "&Ucirc;"); } 
""	{ fprintf (actout, "&Uuml;"); }  
""	{ fprintf (actout, "&Yacute;"); }
""	{ fprintf (actout, "&THORN;"); }
""	{ fprintf (actout, "&szlig;"); }
""	{ fprintf (actout, "&agrave;"); }
""	{ fprintf (actout, "&aacute;"); }
""	{ fprintf (actout, "&acirc;"); } 
""	{ fprintf (actout, "&atilde;"); }  
""	{ fprintf (actout, "&auml;"); }
""	{ fprintf (actout, "&aring;"); }
""	{ fprintf (actout, "&aelig;"); }
""	{ fprintf (actout, "&ccedil;"); }
""	{ fprintf (actout, "&egrave;"); }
""	{ fprintf (actout, "&eacute;"); }
""	{ fprintf (actout, "&ecirc;"); } 
""	{ fprintf (actout, "&euml;"); }  
""	{ fprintf (actout, "&igrave;"); }
""	{ fprintf (actout, "&iacute;"); }
""	{ fprintf (actout, "&icirc;"); } 
""	{ fprintf (actout, "&iuml;"); }  
""	{ fprintf (actout, "&eth;"); }
""	{ fprintf (actout, "&ntilde;"); }
""	{ fprintf (actout, "&ograve;"); }
""	{ fprintf (actout, "&oacute;"); }
""	{ fprintf (actout, "&ocirc;"); } 
""	{ fprintf (actout, "&otilde;"); }  
""	{ fprintf (actout, "&ouml;"); }
	/* ""	{ fprintf (actout, "", yytext); } */
""	{ fprintf (actout, "&oslash;"); }
""	{ fprintf (actout, "&ograve;"); }
""	{ fprintf (actout, "&oacute;"); }
""	{ fprintf (actout, "&ocirc;"); } 
""	{ fprintf (actout, "&ouml;"); }  
""	{ fprintf (actout, "&yacute;"); }
""	{ fprintf (actout, "&thorn;"); }
""	{ fprintf (actout, "&yuml;"); }

\\\"		{ fprintf (actout, "%s", yytext); }
<STRING>\"	{ 
		  fprintf (actout, "%s</FONT>", yytext);
		  switch (oldstate)
		  {
                  case 0: 
		    BEGIN 0; 
		    break;
		  case DEFINELINE: 
		    BEGIN DEFINELINE;
		    oldstate = 0;
		    fprintf (actout, "<FONT COLOR=%s>", definelinecolor);
		    break;
		  case PREPROLINE:
		    BEGIN PREPROLINE;
		    oldstate = 0;
		    fprintf (actout, "<FONT COLOR=%s>", preprolinecolor);
		    break;
		  default:
		    fprintf (stderr, "internal error: bad oldstate\n");
		  }
		}
<STRING>\\\n	{ 
		  fprintf (actout, "%s", yytext);
		  if (oldstate == PREPROLINE) oldstate = 0; 
		}
<STRING>\n	{
		  switch (oldstate)
		  {
		  case 0: 
		    BEGIN STRING;
		    fprintf (actout, "%s", yytext);
		    break;
		  case DEFINELINE:
		    BEGIN STRING;
		    oldstate = 0;	
		    fprintf (actout, "</FONT></STRONG>%s", yytext);
		    fprintf (actout, "<FONT COLOR=%s>", stringcolor);
		    break;
		  case PREPROLINE:
		    BEGIN STRING;
		    oldstate = 0;
		    fprintf (actout, "</FONT>%s", yytext);
		    fprintf (actout, "<FONT COLOR=%s>", stringcolor);	
		    break;
		  default:
		    fprintf (stderr, "internal error: bad oldstate\n");
		  }
		}
<STRING>{stringnospecial}+	{ fprintf (actout, "%s", yytext); }

<ONELINECOMMENT>{onelinecommentnospecial}+	{ 
		  fprintf (actout, "%s", yytext); 
		  BEGIN 0;
		}
<ONELINECOMMENT>\n	{ 
		  fprintf (actout, "%s</FONT>", yytext); 
		  BEGIN 0;
		}
<COMMENT>"/*"	{ fprintf (actout, "%s", yytext); }
<COMMENT>\\\n	{ 
		  fprintf (actout, "%s", yytext);
		  if (oldstate == PREPROLINE) oldstate = 0; 
		}
<COMMENT>\n	{
		  switch (oldstate)
		  {
		  case 0: 
		    BEGIN COMMENT;
		    fprintf (actout, "%s", yytext);
		    break;
		  case DEFINELINE:
		    BEGIN COMMENT;
		    oldstate = 0;	
		    fprintf (actout, "</FONT></STRONG>%s", yytext);
		    fprintf (actout, "<FONT COLOR=%s>", commentcolor);
		    break;
		  case PREPROLINE:
		    BEGIN COMMENT;
		    oldstate = 0;
		    fprintf (actout, "</FONT>%s", yytext);
		    fprintf (actout, "<FONT COLOR=%s>", commentcolor);	
		    break;
		  default:
		    fprintf (stderr, "internal error: bad oldstate\n");
		  }
		}
<COMMENT>{commentnospecial}+	{ fprintf (actout, "%s", yytext); }
<COMMENT>{commentnospecial}/[^/]	{ fprintf (actout, "%s", yytext); }
<COMMENT>"*/"	{
		  fprintf (actout, "%s</FONT>", yytext);
		  switch (oldstate)
		  {
                  case 0: 
		    BEGIN 0; 
		    break;
		  case DEFINELINE: 
		    BEGIN DEFINELINE;
		    oldstate = 0;
		    fprintf (actout, "<FONT COLOR=%s>", definelinecolor);
		    break;
		  case PREPROLINE:
		    BEGIN PREPROLINE;
		    oldstate = 0;
		    fprintf (actout, "<FONT COLOR=%s>", preprolinecolor);
		    break;
		  default:
		    fprintf (stderr, "internal error: bad oldstate\n");
		  }
		}
<COMMENT>{commentnospecial}\/	{ fprintf (actout, "%s", yytext); }

<DEFINELINE>\"		{
		  BEGIN STRING;
		  oldstate = DEFINELINE;
		  fprintf (actout, "</FONT><FONT COLOR=%s>%s"
			   , stringcolor, yytext);
		}
<DEFINELINE>"/*"		{
		  BEGIN COMMENT;
		  oldstate = DEFINELINE;
		  fprintf (actout, "</FONT><FONT COLOR=%s>%s" 
			   , commentcolor, yytext);
		}
<DEFINELINE>\\\n	{ fprintf (actout, "%s", yytext); }
<DEFINELINE>\n	{
		  BEGIN 0;
		  oldstate = 0;
		  fprintf (actout, "</FONT></STRONG>%s", yytext);
		}
<DEFINELINE>{definelinenospecial}+	{ fprintf (actout, "%s", yytext); }
^{ws}("#"|"??=")("define") |
^{ws}("#"|"??=")("undef")	{ 
		  BEGIN DEFINELINE;
		  fprintf (actout, "<STRONG><FONT COLOR=%s>%s"
			   , definelinecolor, yytext); 
		}
<PREPROLINE>\"	{
		  BEGIN STRING;
		  oldstate = PREPROLINE;
		  fprintf (actout, "</FONT><FONT COLOR=%s>%s", 
			   stringcolor, yytext);
		}
<PREPROLINE>"/*"	{
		  BEGIN COMMENT;
		  oldstate = PREPROLINE;
		  fprintf (actout, "</FONT><FONT COLOR=%s>%s", 
			   commentcolor, yytext);
		}
<PREPROLINE>\n	{
		  BEGIN 0;
		  oldstate = 0;
		  fprintf (actout, "</FONT>%s", yytext);
		}
<PREPROLINE>{preprolinenospecial}+	{ fprintf (actout, "%s", yytext); }
^[\t ]*("#"|"??=")("") |
^[\t ]*("#"|"??=")("elif") |
^[\t ]*("#"|"??=")("else") |
^[\t ]*("#"|"??=")("endif") | 
^[\t ]*("#"|"??=")("error") | 
^[\t ]*("#"|"??=")("if ! defined") | 
^[\t ]*("#"|"??=")("if defined") |
^[\t ]*("#"|"??=")("if") |
^[\t ]*("#"|"??=")("ifdef") |
^[\t ]*("#"|"??=")("ifndef") |
^[\t ]*("#"|"??=")("include") |
^[\t ]*("#"|"??=")("line") |
^[\t ]*("#"|"??=")("pragma")	{
		  BEGIN PREPROLINE;
		  fprintf (actout, "<FONT COLOR=%s>%s", 
			   preprolinecolor, yytext);
		  /* highlight preprozessor statements */
		}
^"import"{ws}.*;	{
		  fprintf (actout, "<FONT COLOR=%s>%s</FONT>", 
			   definelinecolor, yytext);
	}

\"	{
	  BEGIN STRING;
	  oldstate = 0;
	  fprintf (actout, "<FONT COLOR=%s>%s", stringcolor, yytext);
	}
"/*"	{
	  BEGIN COMMENT;
	  oldstate = 0;
	  fprintf (actout, "<FONT COLOR=%s>%s", commentcolor, yytext);
	}
"//"{onelinecommentnospecial}*	{
	  BEGIN ONELINECOMMENT;
	  fprintf (actout, "<FONT COLOR=%s>%s", commentcolor, yytext);
	}
[{}]	{
	 fprintf (actout, "%s", yytext);
	}


^{ws}{accessmode}{ws}({id}{wsnl}|"*")*"("({id}|{wsnl}|"\.\.\."|[][*(),])*")"	{
		  fprintf (actout, "<FONT COLOR=%s><STRONG>%s</STRONG></FONT>", keywordcolor, yytext);
		}		  

^({accessmode}{ws}|{ws})"class"{ws}{id}({id}{ws})*	{
		  fprintf (actout, "<FONT COLOR=%s><STRONG>%s</STRONG></FONT>", keywordcolor, yytext);
		}		  

^"typedef"({id}|{ws}|[*{}])*";"	{
		  fprintf (actout, "<FONT COLOR=%s>%s</FONT>", keywordcolor, yytext);
		}		  
"struct "{id}	{
		  fprintf (actout, "<FONT COLOR=%s>%s</FONT>", keywordcolor, yytext);
		}		  
^{ws}{id}":"    {
		  fprintf (actout, "<FONT COLOR=%s><STRONG>%s</STRONG></FONT>", labelcolor, yytext);
		}		  
		  

(char|double|float|int|long|short|signed|void|unsigned)/[^a-zA-Z_0123456789]	{ 
		 fprintf (actout, "%s", yytext); 
		  /* highlight basic types */
		}

(char|double|float|int|long|short|signed|void|unsigned)\*/[^a-zA-Z_0123456789]	{ 
		 fprintf (actout, "%s", yytext); 
		  /* highlight basic pointer types */
		}
(char|double|float|int|long|short|signed|void|unsigned)" *"	{ 
		 fprintf (actout, "%s", yytext); 
		  /* highlight basic pointer types in the common style*/
		}

("auto"|"const"|"extern"|"register"|"static"|"volatile")/[^a-zA-Z_0123456789] 	{ 
		 fprintf (actout, "%s", yytext);
		  /* highlight attributes for types */
		}		  

("break"|"case"|"continue"|"default"|"do"|"else"|"enum"|"for"|"goto"|"if"|"return"|"sizeof"|"struct"|"switch"|"typedef"|"union"|"while"|"try"|"catch"|"throw"|"throws")/[^a-zA-Z_0123456789] 	{ 
		 fprintf (actout, "<FONT COLOR=%s>%s</FONT>", keywordcolor, yytext);
		  /* highlight the rest of the reserved words */
		}
		  
("abort"|"abs"|"acos"|"asctime"|"asin"|"atan"|"atan2"|"atexit"|"atof"|"atoi"|"atol"|"bsearch"|"calloc"|"ceil"|"clearerr"|"clock"|"cos"|"cosh"|"ctime"|"difftime"|"div"|"exit"|"exp"|"fabs"|"fclose"|"feof"|"ferror"|"fflush"|"fgetc"|"fgetpos"|"fgets"|"floor"|"fmod"|"fopen"|"fprintf"|"fputc"|"fputs"|"fread"|"free"|"freopen"|"frexp"|"fscanf"|"fseek"|"fsetpos"|"ftell"|"fwrite"|"getc"|"getchar"|"getenv"|"gets"|"gmtime"|"isalnum"|"isalpha"|"iscntrl"|"isdigit"|"isgraph"|"islower"|"isprint"|"ispunct"|"isspace"|"isupper"|"isxdigit"|"labs"|"ldexp"|"ldiv"|"localtime"|"log"|"log10"|"longjmp"|"malloc"|"memchr"|"memcmp"|"memcpy"|"memmove"|"memset"|"mktime"|"modf"|"perror"|"pow"|"printf"|"putc"|"putchar"|"puts"|"qsort"|"raise"|"rand"|"realloc"|"remove"|"rename"|"rewind"|"scanf"|"setbuf"|"setvbuf"|"setjmp"|"signal"|"sin"|"sinh"|"sprintf"|"sqrt"|"srand"|"sscanf"|"strcat"|"strchr"|"strcmp"|"strcpy"|"strcspn"|"strerror"|"strftime"|"strlen"|"strncat"|"strncmp"|"strncpy"|"strpbrk"|"strrchr"|"strspn"|"strstr"|"strtod"|"strtok"|"strtol"|"strtul"|"system"|"tan"|"tanh"|"time"|"tmpfile"|"tmpnam"|"tolower"|"toupper"|"ungetc"|"vfprintf"|"vprintf"|"vsprintf")/[^a-zA-Z_0123456789]	{
		 fprintf (actout, "%s",yytext);
		  /* highlight lybrary functions */
		}

("close"|"create"|"fstat"|"lseek"|"open"|"read"|"abrk"|"stat"|"unlink"|"write")/[^a-zA-Z_0123456789]	{
		 fprintf (actout, "%s",yytext);
		  /* highlight basic system calls */
		}

[\t ]+		{ 
		 fprintf (actout, "%s", yytext); 
		  /* let whitespaces like they are */
		}

[a-zA-Z_][a-zA-Z_0123456789]*	{
		 fprintf (actout, "%s",yytext);
		}

.		{ 
		 fprintf (actout, "%s",yytext); 
		}
%%

int insert (FILE * outfile, char *filename);
/* 
 * inserts the file "filename" in the current directory (or if given the 
 * complete path this file) into the output file.
 */

static char *prog;
static void usage() 
{
  fprintf (stderr, 
           "usage: %s -V [-t title] [-w width] [file_to_convert ...]\n", 
           prog);
  exit(3);
}

int
main(int argc, char *argv[])
{
  char *outfilename = (char *) NULL;
  int  i            = 0;
  int  ch           = 0;
  int  width        = 80;
  int  rc           = 0;
  char *title       = (char *) NULL;
  char *pt          = (char *) NULL;

  char gzipcmd[10];

  pt = strrchr(argv[0], '/');
  prog = (pt)? ++pt : argv[0]; /* basename */
  while (-1 != (ch = getopt(argc, argv, "h?t:w:V")))
    {
      switch(ch) 
	{
	case 't':
	  title = optarg;
	  break;
	case 'w':
	  if (1 != sscanf(optarg, "%d", &width))
	    { usage(); }
	  break;
	case 'V':
	  fprintf(stderr, "java2html version %s%s\n",
		  VERSION, 
                  COMPRESSION ? "" : " (no compression)");
	  exit(0);
	default:
	  usage();
	}
    }
  argc -= optind;
  argv += optind;

  if (0 == argc) 
    {
      actin  = stdin;
      actout = stdout;
      
      if ((pt = getenv("PATH_TRANSLATED")) 
	  && getenv("GATEWAY_INTERFACE")) 
	{
	  /* CGI */
	  if (!title)
	    { title = pt; }
	  actin = fopen (pt, "r");
	  if (!actin) 
	    {
	      fprintf (actout, "Content-Type: text/html\n\n");
	      fprintf (actout, "<HTML><HEAD><TITLE>error in CGI '%s'", prog);
	      fprintf (actout, "</TITLE></HEAD>\n");
	      fprintf (actout, "<BODY><H1>error in CGI program '%s': ", prog);
	      fprintf (actout, "cannot read file '%s'</H1></BODY>\n", pt);
	      fprintf (actout, "</HTML>\n");
	      exit(1);
	    }
	  fprintf (actout, "Content-Type: text/html\n");
          if (COMPRESSION)
	    {
	      /* should we compress output with gzip? */
	      do 
		{
		  if (!(pt = getenv("HTTP_ACCEPT_ENCODING")))
		    { break; } /* no header Accept-Encoding: */
		  if (!strstr(pt, "gzip"))
		    { break; } /* browser doesn't understand gzip format */
		  if ((pt = getenv("REMOTE_ADDR")) 
		      && (!strncmp(pt, "127", 3)))
		    { break; } /* never compress to local client */
		  /* OK, compress output with gzip */
		  fprintf(actout, "Content-Encoding: x-gzip\n\n");
		  fflush(actout);
		  sprintf(gzipcmd, "gzip -%d", COMPRESSION);
		  if (!(actout = (FILE *) popen(gzipcmd, "w")))
		    { exit(4); } /* cannot fork, gzip not found, ... */
		} 
	      while (0);
	    }
	} /* end of CGI section */
      fprintf (actout, "\n<HTML>\n<HEAD>\n");
      fprintf (actout, "<TITLE>%s", title ? title : "stdin");
      fprintf (actout, "</TITLE>\n</HEAD>\n\n");
      fprintf (actout, "<BODY BGCOLOR=\"%s\">\n", bgcolor);
      (void) insert (actout, ownhead);
      fprintf (actout, "<PRE WIDTH=\"%d\">", width);
  
      yyin  = actin;
      yyout = actout;

      BEGIN 0;
      yylex();
      
      fprintf (actout,"</PRE>\n</BODY>\n\n</HTML>\n");
    } 
  else 
    for (i = 0; i < argc; i++) 
      {
	actin = fopen (argv[i], "r");
	if (!actin) 
	  {
	    fprintf (stderr, "%s: cannot read file '%s'\n", prog, argv[i]);
	    rc = rc < 1 ? 1 : rc;
	    continue;
	  }
      
	outfilename = malloc (sizeof (char)*strlen (argv[i]) + 6);
	strcat (strcpy (outfilename, argv[i]), ".html");
	actout = fopen (outfilename, "w");
	if (!actout) 
	  {
	    fprintf (stderr, "%s: cannot write file '%s'\n", prog, outfilename);
	    rc = rc < 2 ? 2 : rc;
	    continue;
	  }
	outfilename[strlen(outfilename) - 5] = '\0'; /* chop '.html' suffix */
	fprintf (actout, "<HTML>\n<HEAD>\n");
	fprintf (actout, "<TITLE>%s", outfilename);
	fprintf (actout, "</TITLE>\n</HEAD>\n\n");
	fprintf (actout, "<BODY BGCOLOR=\"%s\">\n", bgcolor);
	(void) insert (actout, ownhead);
	fprintf (actout, "<PRE WIDTH=\"%d\">", width);

	yyin  = actin;
	yyout = actout;

	BEGIN 0;
	yylex ();

	fprintf (actout, "</PRE>\n");
	(void) insert (actout, ownbottom);
	fprintf (actout, "</BODY>\n\n</HTML>\n");
	free (outfilename);
	fclose (actin);
	fclose (actout);
      }
  return rc;
  /* unreachable, but suppresses compiler warning. */
  yyunput(0, NULL);
}

int
insert (FILE * outfile, char *filename)
{
  int c;
  FILE * infile;
 
  infile = fopen (filename, "r");
  if (NULL != infile)
    {
      c = fgetc (infile);
      while (EOF != c)
        { 
          fputc (c, outfile);
          c = fgetc (infile);
        }
      fclose (infile);
      return 1;
    }
  else
    { return 0; }
}
